/* $XFree86: xc/programs/Xserver/hw/xfree86/drivers/MQ/MQ_shadow.c,v 1.3 2000/02/08 13:13:18 eich Exp $ */

/*
   Copyright (c) 1999,  The XFree86 Project Inc. 
   Written by Mark Vojkovich <markv@valinux.com>
*/

#include "xf86.h"
#include "xf86_OSproc.h"
#include "xf86Resources.h"
#include "xf86_ansic.h"
#include "xf86PciInfo.h"
#include "xf86Pci.h"
#include "mq2hw.h"
#include "mq.h"
#include "shadowfb.h"
#include "servermd.h"



void
MQRefreshArea(ScrnInfoPtr pScrn, int num, BoxPtr pbox)
{
    MQPtr pMQ = MQPTR(pScrn);
    int width, height, Bpp, FBPitch;
    unsigned char *src, *dst;
   
    Bpp = pScrn->bitsPerPixel >> 3;
    FBPitch = BitmapBytePad(pScrn->displayWidth * pScrn->bitsPerPixel);

    while(num--)
    {
	width = (pbox->x2 - pbox->x1) * Bpp;
	height = pbox->y2 - pbox->y1;
	src = pMQ->ShadowPtr + (pbox->y1 * pMQ->ShadowPitch) + 
						(pbox->x1 * Bpp);
	dst = pMQ->FbStart + (pbox->y1 * FBPitch) + (pbox->x1 * Bpp);

	while(height--)
	{
	    memcpy(dst, src, width);
	    dst += FBPitch;
	    src += pMQ->ShadowPitch;
	}
	
	pbox++;
    }
} 

void
MQPointerMoved(int index, int x, int y)
{
    ScrnInfoPtr pScrn = xf86Screens[index];
    MQPtr pMQ = MQPTR(pScrn);
    int newX, newY;

    if(pMQ->Rotate == 1)
    {
	newX = pScrn->pScreen->height - y - 1;
	newY = x;
    }
    else
    {
	newX = y;
	newY = pScrn->pScreen->width - x - 1;
    }

    (*pMQ->PointerMoved)(index, newX, newY);
}

void
MQRefreshArea8(ScrnInfoPtr pScrn, int num, BoxPtr pbox)
{
    MQPtr pMQ = MQPTR(pScrn);
    int count, width, height, y1, y2, dstPitch, srcPitch;
    CARD8 *dstPtr, *srcPtr, *src;
    CARD32 *dst;

    dstPitch = pScrn->displayWidth;
    srcPitch = -pMQ->Rotate * pMQ->ShadowPitch;

    while(num--)
    {
	width = pbox->x2 - pbox->x1;
	y1 = pbox->y1 & ~3;
	y2 = (pbox->y2 + 3) & ~3;
	height = (y2 - y1) >> 2;  /* in dwords */

	if(pMQ->Rotate == 1)
	{
	    dstPtr = pMQ->FbStart + 
			(pbox->x1 * dstPitch) + pScrn->virtualX - y2;
	    srcPtr = pMQ->ShadowPtr + ((1 - y2) * srcPitch) + pbox->x1;
	}
	else
	{
	    dstPtr = pMQ->FbStart + 
			((pScrn->virtualY - pbox->x2) * dstPitch) + y1;
	    srcPtr = pMQ->ShadowPtr + (y1 * srcPitch) + pbox->x2 - 1;
	}

	while(width--)
	{
	    src = srcPtr;
	    dst = (CARD32*)dstPtr;
	    count = height;
	    while(count--)
	    {
		*(dst++) = src[0] | (src[srcPitch] << 8) | 
					(src[srcPitch * 2] << 16) | 
					(src[srcPitch * 3] << 24);
		src += srcPitch * 4;
	    }
	    srcPtr += pMQ->Rotate;
	    dstPtr += dstPitch;
	}

	pbox++;
    }
} 


void
MQRefreshArea16(ScrnInfoPtr pScrn, int num, BoxPtr pbox)
{
    MQPtr pMQ = MQPTR(pScrn);
    int count, width, height, y1, y2, dstPitch, srcPitch;
    CARD16 *dstPtr, *srcPtr, *src;
    CARD32 *dst;

    dstPitch = pScrn->displayWidth;
    srcPitch = -pMQ->Rotate * pMQ->ShadowPitch >> 1;

    while(num--)
    {
	width = pbox->x2 - pbox->x1;
	y1 = pbox->y1 & ~1;
	y2 = (pbox->y2 + 1) & ~1;
	height = (y2 - y1) >> 1;  /* in dwords */

	if(pMQ->Rotate == 1)
	{
	    dstPtr = (CARD16*)pMQ->FbStart + 
			(pbox->x1 * dstPitch) + pScrn->virtualX - y2;
	    srcPtr = (CARD16*)pMQ->ShadowPtr + 
			((1 - y2) * srcPitch) + pbox->x1;
	}
	else
	{
	    dstPtr = (CARD16*)pMQ->FbStart + 
			((pScrn->virtualY - pbox->x2) * dstPitch) + y1;
	    srcPtr = (CARD16*)pMQ->ShadowPtr + 
			(y1 * srcPitch) + pbox->x2 - 1;
	}

	while(width--)
	{
	    src = srcPtr;
	    dst = (CARD32*)dstPtr;
	    count = height;
	    while(count--)
	    {
		*(dst++) = src[0] | (src[srcPitch] << 16);
		src += srcPitch * 2;
	    }
	    srcPtr += pMQ->Rotate;
	    dstPtr += dstPitch;
	}

	pbox++;
    }
}


/* this one could be faster */
void
MQRefreshArea24(ScrnInfoPtr pScrn, int num, BoxPtr pbox)
{
    MQPtr pMQ = MQPTR(pScrn);
    int count, width, height, y1, y2, dstPitch, srcPitch;
    CARD8 *dstPtr, *srcPtr, *src;
    CARD32 *dst;

    dstPitch = BitmapBytePad(pScrn->displayWidth * 24);
    srcPitch = -pMQ->Rotate * pMQ->ShadowPitch;

    while(num--)
    {
        width = pbox->x2 - pbox->x1;
        y1 = pbox->y1 & ~3;
        y2 = (pbox->y2 + 3) & ~3;
        height = (y2 - y1) >> 2;  /* blocks of 3 dwords */

	if(pMQ->Rotate == 1)
	{
	    dstPtr = pMQ->FbStart + 
			(pbox->x1 * dstPitch) + ((pScrn->virtualX - y2) * 3);
	    srcPtr = pMQ->ShadowPtr + ((1 - y2) * srcPitch) + (pbox->x1 * 3);
	}
	else
	{
	    dstPtr = pMQ->FbStart + 
			((pScrn->virtualY - pbox->x2) * dstPitch) + (y1 * 3);
	    srcPtr = pMQ->ShadowPtr + (y1 * srcPitch) + (pbox->x2 * 3) - 3;
	}

	while(width--)
	{
	    src = srcPtr;
	    dst = (CARD32*)dstPtr;
	    count = height;
	    while(count--)
	    {
		dst[0] = src[0] | (src[1] << 8) | (src[2] << 16) |
				(src[srcPitch] << 24);		
		dst[1] = src[srcPitch + 1] | (src[srcPitch + 2] << 8) |
				(src[srcPitch * 2] << 16) |
				(src[(srcPitch * 2) + 1] << 24);		
		dst[2] = src[(srcPitch * 2) + 2] | (src[srcPitch * 3] << 8) |
				(src[(srcPitch * 3) + 1] << 16) |
				(src[(srcPitch * 3) + 2] << 24);	
		dst += 3;
		src += srcPitch * 4;
	    }
	    srcPtr += pMQ->Rotate * 3;
	    dstPtr += dstPitch; 
	}

	pbox++;
    }
}

void
MQRefreshArea32(ScrnInfoPtr pScrn, int num, BoxPtr pbox)
{
    MQPtr pMQ = MQPTR(pScrn);
    int count, width, height, dstPitch, srcPitch;
    CARD32 *dstPtr, *srcPtr, *src, *dst;

    dstPitch = pScrn->displayWidth;
    srcPitch = -pMQ->Rotate * pMQ->ShadowPitch >> 2;

    while(num--)
    {
	width = pbox->x2 - pbox->x1;
	height = pbox->y2 - pbox->y1;

	if(pMQ->Rotate == 1)
	{
	    dstPtr = (CARD32*)pMQ->FbStart + 
			(pbox->x1 * dstPitch) + pScrn->virtualX - pbox->y2;
	    srcPtr = (CARD32*)pMQ->ShadowPtr + 
			((1 - pbox->y2) * srcPitch) + pbox->x1;
	}
	else
	{
	    dstPtr = (CARD32*)pMQ->FbStart + 
			((pScrn->virtualY - pbox->x2) * dstPitch) + pbox->y1;
	    srcPtr = (CARD32*)pMQ->ShadowPtr + 
			(pbox->y1 * srcPitch) + pbox->x2 - 1;
	}

	while(width--)
	{
	    src = srcPtr;
	    dst = dstPtr;
	    count = height;
	    while(count--)
	    {
		*(dst++) = *src;
		src += srcPitch;
	    }
	    srcPtr += pMQ->Rotate;
	    dstPtr += dstPitch;
	}

	pbox++;
    }
}
